import attr
import datetime
import aiohttp
import random
import re
import json
import os
import time
import errno
import string
from yarl import URL
from http.cookies import SimpleCookie, BaseCookie

# TODO: Only import when required
# Or maybe just replace usage with `html.parser`?
import bs4

from ._common import log, kw_only
from . import _graphql, _util, _exception

from typing import Optional, Mapping, Callable, Any, Awaitable

SERVER_JS_DEFINE_REGEX = re.compile(r'(?:'
                                    r'\(new ServerJS\(\)\)(?:;s)?'
                                    r'|\(require\("ServerJS(?:Define)?"\)\)\(\)'
                                    r').handle(?:Defines)?\(')
SERVER_JS_DEFINE_JSON_DECODER = json.JSONDecoder()


def write_html_to_temp(html: str) -> str:
    random_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=6))
    file_path = f"/tmp/fbchat-debug/serverjsdefine-{int(time.time())}-{random_id}"
    try:
        os.makedirs(os.path.dirname(file_path))
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
    with open(file_path, "w") as file:
        file.write(html)
    return file_path


def parse_server_js_define(html: str) -> Mapping[str, Any]:
    """Parse ``ServerJSDefine`` entries from a HTML document."""
    # Find points where we should start parsing
    define_splits = SERVER_JS_DEFINE_REGEX.split(html)

    # TODO: Extract jsmods "require" and "define" from `bigPipe.onPageletArrive`?

    # Skip leading entry
    _, *define_splits = define_splits

    if not define_splits:
        file_name = write_html_to_temp(html)
        raise _exception.ParseError("Could not find any ServerJSDefine", data_file=file_name)
    if len(define_splits) > 2:
        file_name = write_html_to_temp(html)
        raise _exception.ParseError("Found too many ServerJSDefine", data_file=file_name)
    try:
        parsed, _ = SERVER_JS_DEFINE_JSON_DECODER.raw_decode(define_splits[0], idx=0)
    except json.JSONDecodeError as e:
        file_name = write_html_to_temp(html)
        raise _exception.ParseError("Invalid ServerJSDefine: not json", data_file=file_name) from e
    try:
        rtn = parsed["define"]
    except KeyError:
        file_name = write_html_to_temp(html)
        raise _exception.ParseError("Invalid ServerJSDefine: missing define key", data_file=file_name)

    if not isinstance(rtn, list):
        file_name = write_html_to_temp(html)
        raise _exception.ParseError("Invalid ServerJSDefine: define value is not a list", data_file=file_name)

    # Convert to a dict
    return _util.get_jsmods_define(rtn)


def base36encode(number: int) -> str:
    """Convert from Base10 to Base36."""
    # Taken from https://en.wikipedia.org/wiki/Base36#Python_implementation
    chars = "0123456789abcdefghijklmnopqrstuvwxyz"

    sign = "-" if number < 0 else ""
    number = abs(number)
    result = ""

    while number > 0:
        number, remainder = divmod(number, 36)
        result = chars[remainder] + result

    return sign + result


def prefix_url(url: str) -> str:
    if url.startswith("/"):
        return "https://www.messenger.com" + url
    return url


def generate_message_id(now: datetime.datetime, client_id: str) -> str:
    k = _util.datetime_to_millis(now)
    l = int(random.random() * 4294967295)
    return "<{}:{}-{}@mail.projektitan.com>".format(k, l, client_id)


def get_user_id(session: aiohttp.ClientSession) -> str:
    try:
        rtn = session.cookie_jar.filter_cookies(URL("https://messenger.com")).get("c_user")
    except (AttributeError, KeyError):
        raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
    if rtn is None:
        raise _exception.ParseError("Could not find user id", data=session.cookie_jar._cookies)
    return rtn if isinstance(rtn, str) else str(rtn.value)


def session_factory() -> aiohttp.ClientSession:
    from . import __version__

    session = aiohttp.ClientSession(headers={
        "Referer": "https://www.messenger.com/",
        # We won't try to set a fake user agent to mask our presence!
        # Facebook allows us access anyhow, and it makes our motives clearer:
        # We're not trying to cheat Facebook, we simply want to access their service
        "User-Agent": "fbchat-asyncio/{}".format(__version__)
    })
    return session


def login_cookies(at: datetime.datetime):
    return {"act": "{}/0".format(_util.datetime_to_millis(at))}


def client_id_factory() -> str:
    return hex(int(random.random() * 2 ** 31))[2:]


def find_form_request(html: str):
    soup = bs4.BeautifulSoup(html, "html.parser", parse_only=bs4.SoupStrainer("form"))

    form = soup.form
    if not form:
        raise _exception.ParseError("Could not find form to submit", data=html)

    url = form.get("action")
    if not url:
        raise _exception.ParseError("Could not find url to submit to", data=form)

    # From what I've seen, it'll always do this!
    if url.startswith("/"):
        url = "https://www.facebook.com" + url

    # It's okay to set missing values to something crap, the values are localized, and
    # hence are not available in the raw HTML
    data = {
        x["name"]: x.get("value", "[missing]")
        for x in form.find_all(["input", "button"])
    }
    return url, data


async def two_factor_helper(session: aiohttp.ClientSession, r: aiohttp.ClientResponse,
                            on_2fa_callback: Callable[[], Awaitable[int]]) -> str:
    url, data = find_form_request(await r.text())

    # You don't have to type a code if your device is already saved
    # Repeats if you get the code wrong
    while "approvals_code" in data:
        data["approvals_code"] = await on_2fa_callback()
        log.info("Submitting 2FA code")
        r = await session.post(url, data=data, allow_redirects=False,
                               cookies=login_cookies(_util.now()))
        log.debug("2FA location: %s", r.headers.get("Location"))
        url, data = find_form_request(await r.text())

    # TODO: Can be missing if checkup flow was done on another device in the meantime?
    if "name_action_selected" in data:
        data["name_action_selected"] = "save_device"
        log.info("Saving browser")
        r = await session.post(url, data=data, allow_redirects=False,
                               cookies=login_cookies(_util.now()))
        log.debug("2FA location: %s", r.headers.get("Location"))
        url = r.headers.get("Location")
        if url and url.startswith("https://www.messenger.com/login/auth_token/"):
            return url
        url, data = find_form_request(await r.text())

    log.info("Starting Facebook checkup flow")
    r = await session.post(url, data=data, allow_redirects=False,
                           cookies=login_cookies(_util.now()))
    log.debug("2FA location: %s", r.headers.get("Location"))

    url, data = find_form_request(await r.text())
    if "verification_method" in data:
        raise _exception.NotLoggedIn(
            "Your account is locked, and you need to log in using a browser, and verify it there!"
        )
    if "submit[This was me]" not in data or "submit[This wasn't me]" not in data:
        raise _exception.ParseError("Could not fill out form properly (2)", data=data)
    data["submit[This was me]"] = "[any value]"
    del data["submit[This wasn't me]"]
    log.info("Verifying login attempt")

    r = await session.post(url, data=data, allow_redirects=False,
                           cookies=login_cookies(_util.now()))
    log.debug("2FA location: %s", r.headers.get("Location"))

    url, data = find_form_request(await r.text())
    if "name_action_selected" not in data:
        raise _exception.ParseError("Could not fill out form properly (3)", data=data)
    data["name_action_selected"] = "save_device"
    log.info("Saving device again")

    r = await session.post(url, data=data, allow_redirects=False,
                           cookies=login_cookies(_util.now()))
    log.debug("2FA location: %s", r.headers.get("Location"))
    return r.headers.get("Location")


def get_error_data(html: str) -> Optional[str]:
    """Get error message from a request."""
    soup = bs4.BeautifulSoup(
        html, "html.parser", parse_only=bs4.SoupStrainer("form", id="login_form")
    )
    # Attempt to extract and format the error string
    # The error message is in the user's own language!
    return " ".join(list(soup.stripped_strings)[1:3]) or None


def get_fb_dtsg(define) -> Optional[str]:
    if "DTSGInitData" in define:
        return define["DTSGInitData"]["token"]
    elif "DTSGInitialData" in define:
        return define["DTSGInitialData"]["token"]
    return None


@attr.s(slots=True, kw_only=kw_only, repr=False, eq=False, auto_attribs=True)
class Session:
    """Stores and manages state required for most Facebook requests.

    This is the main class, which is used to login to Facebook.
    """

    _user_id: str
    _fb_dtsg: str
    _revision: int
    _session: aiohttp.ClientSession = attr.ib(factory=session_factory)
    _counter: int = 0
    _client_id: str = attr.ib(factory=client_id_factory)

    @property
    def user(self):
        """The logged in user."""
        from . import _threads

        # TODO: Consider caching the result

        return _threads.User(session=self, id=self._user_id)

    def __repr__(self) -> str:
        # An alternative repr, to illustrate that you can't create the class directly
        return "<fbchat.Session user_id={}>".format(self._user_id)

    def _get_params(self):
        self._counter += 1  # TODO: Make this operation atomic / thread-safe
        return {
            "__a": 1,
            "__req": base36encode(self._counter),
            "__rev": self._revision,
            "fb_dtsg": self._fb_dtsg,
        }

    # TODO: Add ability to load previous cookies in here, to avoid 2fa flow
    @classmethod
    async def login(
        cls, email: str, password: str, on_2fa_callback: Callable[[], Awaitable[int]] = None
    ):
        """Login the user, using ``email`` and ``password``.

        Args:
            email: Facebook ``email``, ``id`` or ``phone number``
            password: Facebook account password
            on_2fa_callback: Function that will be called, in case a two factor
                authentication code is needed. This should return the requested code.

                Tested using SMS and authentication applications. If you have both
                enabled, you might not receive an SMS code, and you'll have to use the
                authentication application.

                Note: Facebook limits the amount of codes they will give you, so if you
                don't receive a code, be patient, and try again later!

        Example:
            >>> import fbchat
            >>> import getpass
            >>> session = fbchat.Session.login(
            ...     input("Email: "),
            ...     getpass.getpass(),
            ...     on_2fa_callback=lambda: input("2FA Code: ")
            ... )
            Email: abc@gmail.com
            Password: ****
            2FA Code: 123456
            >>> session.user.id
            "1234"
        """
        session = session_factory()

        data = {
            # "jazoest": "2754",
            # "lsd": "AVqqqRUa",
            "initial_request_id": "x",  # any, just has to be present
            # "timezone": "-120",
            # "lgndim": "eyJ3IjoxNDQwLCJoIjo5MDAsImF3IjoxNDQwLCJhaCI6ODc3LCJjIjoyNH0=",
            # "lgnrnd": "044039_RGm9",
            "lgnjs": "n",
            "email": email,
            "pass": password,
            "login": "1",
            "persistent": "1",  # Changes the cookie type to have a long "expires"
            "default_persistent": "0",
        }

        try:
            # Should hit a redirect to https://www.messenger.com/
            # If this does happen, the session is logged in!
            r = await session.post(
                "https://www.messenger.com/login/password/",
                data=data,
                allow_redirects=False,
                cookies=login_cookies(_util.now()),
            )
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        url = r.headers.get("Location")

        # We weren't redirected, hence the email or password was wrong
        if not url:
            error = get_error_data(await r.text())
            raise _exception.NotLoggedIn(error)

        if "checkpoint" in url:
            if not on_2fa_callback:
                raise _exception.NotLoggedIn(
                    "2FA code required! Please supply `on_2fa_callback` to .login"
                )
            # Get a facebook.com/checkpoint/start url that handles the 2FA flow
            # This probably works differently for Messenger-only accounts
            url = _util.get_url_parameter(url, "next")
            if not url.startswith("https://www.facebook.com/checkpoint/start/"):
                raise _exception.ParseError("Failed 2fa flow (1)", data=url)

            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
            url = r.headers.get("Location")
            if not url or not url.startswith("https://www.facebook.com/checkpoint/"):
                raise _exception.ParseError("Failed 2fa flow (2)", data=url)

            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
            url = await two_factor_helper(session, r, on_2fa_callback)

            if not url.startswith("https://www.messenger.com/login/auth_token/"):
                raise _exception.ParseError("Failed 2fa flow (3)", data=url)

            r = await session.get(url, allow_redirects=False, cookies=login_cookies(_util.now()))
            url = r.headers.get("Location")

        if url != "https://www.messenger.com/":
            error = get_error_data(await r.text())
            raise _exception.NotLoggedIn("Failed logging in: {}, {}".format(url, error))

        try:
            return await cls._from_session(session=session)
        except _exception.NotLoggedIn as e:
            raise _exception.ParseError("Failed loading session", data=r) from e

    async def is_logged_in(self) -> bool:
        """Send a request to Facebook to check the login status.

        Returns:
            Whether the user is still logged in

        Example:
            >>> assert session.is_logged_in()
        """
        # Send a request to the login url, to see if we're directed to the home page
        try:
            r = await self._session.get(prefix_url("/login/"), allow_redirects=False)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)
        return "https://www.messenger.com/" == r.headers.get("Location")

    async def logout(self) -> None:
        """Safely log out the user.

        The session object must not be used after this action has been performed!

        Example:
            >>> session.logout()
        """
        data = {"fb_dtsg": self._fb_dtsg}
        try:
            r = await self._session.post(
                prefix_url("/logout/"), data=data, allow_redirects=False
            )
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        if "Location" not in r.headers:
            raise _exception.FacebookError("Failed logging out, was not redirected!")
        if "https://www.messenger.com/login/" != r.headers["Location"]:
            raise _exception.FacebookError(
                "Failed logging out, got bad redirect: {}".format(r.headers["Location"])
            )

    @classmethod
    async def _from_session(cls, session):
        # TODO: Automatically set user_id when the cookie changes in the session
        user_id = get_user_id(session)

        # Make a request to the main page to retrieve ServerJSDefine entries
        try:
            r = await session.get(prefix_url("/"), allow_redirects=False)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)

        define = parse_server_js_define(await r.text())

        fb_dtsg = get_fb_dtsg(define)
        if fb_dtsg is None:
            raise _exception.ParseError("Could not find fb_dtsg", data=define)
        if not fb_dtsg:
            # Happens when the client is not actually logged in
            raise _exception.NotLoggedIn(
                "Found empty fb_dtsg, the session was probably invalid."
            )

        try:
            revision = int(define["SiteData"]["client_revision"])
        except TypeError:
            raise _exception.ParseError("Could not find client revision", data=define)

        return cls(user_id=user_id, fb_dtsg=fb_dtsg, revision=revision, session=session)

    def get_cookies(self) -> Optional[Mapping[str, str]]:
        """Retrieve session cookies, that can later be used in `from_cookies`.

        Returns:
            A dictionary containing session cookies

        Example:
            >>> cookies = session.get_cookies()
        """
        cookie = self._session.cookie_jar.filter_cookies(URL("https://messenger.com"))
        return {key: morsel.value for key, morsel in cookie.items()}

    @classmethod
    async def from_cookies(cls, cookies: Mapping[str, str]):
        """Load a session from session cookies.

        Args:
            cookies: A dictionary containing session cookies

        Example:
            >>> cookies = session.get_cookies()
            >>> # Store cookies somewhere, and then subsequently
            >>> session = fbchat.Session.from_cookies(cookies)
        """
        session = session_factory()

        if isinstance(cookies, BaseCookie):
            cookie = cookies
        else:
            cookie = SimpleCookie()
            for key, value in cookies.items():
                cookie[key] = value
                cookie[key].update({"domain": "messenger.com", "path": "/"})
        session.cookie_jar.update_cookies(cookie, URL("https://messenger.com"))

        return await cls._from_session(session=session)

    async def _post(self, url, data, files=None, as_graphql=False):
        data.update(self._get_params())
        if files:
            payload = aiohttp.FormData()
            for key, value in data.items():
                payload.add_field(key, str(value))
            for key, (name, file, content_type) in files.items():
                payload.add_field(key, file, filename=name, content_type=content_type)
            data = payload
        try:
            r = await self._session.post(prefix_url(url), data=data)
        except aiohttp.ClientError as e:
            _exception.handle_requests_error(e)
            raise Exception("handle_requests_error did not raise exception")
        _exception.handle_http_error(r.status)
        text = await r.text()
        if text is None or len(text) == 0:
            raise _exception.HTTPError("Error when sending request: Got empty response")
        if as_graphql:
            return _graphql.response_to_json(text)
        else:
            text = _util.strip_json_cruft(text)
            j = _util.parse_json(text)
            log.debug(j)
            return j

    async def _payload_post(self, url, data, files=None):
        j = await self._post(url, data, files=files)
        _exception.handle_payload_error(j)

        # update fb_dtsg token if received in response
        if "jsmods" in j:
            define = _util.get_jsmods_define(j["jsmods"]["define"])
            fb_dtsg = get_fb_dtsg(define)
            if fb_dtsg:
                self._fb_dtsg = fb_dtsg

        try:
            return j["payload"]
        except (KeyError, TypeError) as e:
            raise _exception.ParseError("Missing payload", data=j) from e

    async def _graphql_requests(self, *queries):
        # TODO: Explain usage of GraphQL, probably in the docs
        # Perhaps provide this API as public?
        data = {
            "method": "GET",
            "response_format": "json",
            "queries": _graphql.queries_to_json(*queries),
        }
        return await self._post("/api/graphqlbatch/", data, as_graphql=True)

    async def _do_send_request(self, data):
        now = _util.now()
        offline_threading_id = _util.generate_offline_threading_id()
        data["client"] = "mercury"
        data["author"] = "fbid:{}".format(self._user_id)
        data["timestamp"] = _util.datetime_to_millis(now)
        data["source"] = "source:chat:web"
        data["offline_threading_id"] = offline_threading_id
        data["message_id"] = offline_threading_id
        data["threading_id"] = generate_message_id(now, self._client_id)
        data["ephemeral_ttl_mode:"] = "0"
        j = await self._post("/messaging/send/", data)

        _exception.handle_payload_error(j)

        try:
            message_ids = [
                (action["message_id"], action["thread_fbid"])
                for action in j["payload"]["actions"]
                if "message_id" in action
            ]
            if len(message_ids) != 1:
                log.warning("Got multiple message ids' back: {}".format(message_ids))
            return message_ids[0]
        except (KeyError, IndexError, TypeError) as e:
            raise _exception.ParseError("No message IDs could be found", data=j) from e



import attr
import datetime
import enum
from .._common import attrs_default
from .. import _util

from typing import Optional


class ThreadLocation(enum.Enum):
    """Used to specify where a thread is located (inbox, pending, archived, other)."""

    INBOX = "INBOX"
    PENDING = "PENDING"
    ARCHIVED = "ARCHIVED"
    OTHER = "OTHER"

    @classmethod
    def _parse(cls, value: str):
        return cls(value.lstrip("FOLDER_"))


@attrs_default
class ActiveStatus:
    #: Whether the user is active now
    active: bool
    #: When the user was last active
    last_active: Optional[datetime.datetime] = None
    #: Whether the user is playing Messenger game now
    in_game: Optional[bool] = None

    @classmethod
    def _from_orca_presence(cls, data):
        # TODO: Handle `c` and `vc` keys (Probably some binary data)
        return cls(
            active=data["p"] in [2, 3],
            last_active=_util.seconds_to_datetime(data["l"]) if "l" in data else None,
            in_game=None,
        )


@attrs_default
class Image:
    #: URL to the image
    url: str
    #: Width of the image
    width: Optional[int] = None
    #: Height of the image
    height: Optional[int] = None

    @classmethod
    def _from_uri(cls, data):
        return cls(
            url=data["uri"],
            width=int(data["width"]) if data.get("width") else None,
            height=int(data["height"]) if data.get("height") else None,
        )

    @classmethod
    def _from_url(cls, data):
        return cls(
            url=data["url"],
            width=int(data["width"]) if data.get("width") else None,
            height=int(data["height"]) if data.get("height") else None,
        )

    @classmethod
    def _from_uri_or_none(cls, data):
        if data is None:
            return None
        if data.get("uri") is None:
            return None
        return cls._from_uri(data)

    @classmethod
    def _from_url_or_none(cls, data):
        if data is None:
            return None
        if data.get("url") is None:
            return None
        return cls._from_url(data)



import attr
import datetime
from . import Image, Attachment
from .._common import attrs_default
from .. import _util, _exception

from typing import Optional


@attrs_default
class LocationAttachment(Attachment):
    """Represents a user location.

    Latitude and longitude OR address is provided by Facebook.
    """

    #: Latitude of the location
    latitude: Optional[float] = None
    #: Longitude of the location
    longitude: Optional[float] = None
    #: Image showing the map of the location
    image: Optional[Image] = None
    #: URL to Bing maps with the location
    url: Optional[str] = None
    # Address of the location
    address: Optional[str] = None

    @classmethod
    def _from_graphql(cls, data):
        url = data.get("url")
        address = _util.get_url_parameter(_util.get_url_parameter(url, "u"), "where1")
        if not address:
            raise _exception.ParseError("Could not find location address", data=data)
        try:
            latitude, longitude = [float(x) for x in address.split(", ")]
            address = None
        except ValueError:
            latitude, longitude = None, None

        return cls(
            id=int(data["deduplication_key"]),
            latitude=latitude,
            longitude=longitude,
            image=Image._from_uri_or_none(data["media"].get("image"))
            if data.get("media")
            else None,
            url=url,
            address=address,
        )


@attrs_default
class LiveLocationAttachment(LocationAttachment):
    """Represents a live user location."""

    #: Name of the location
    name: Optional[str] = None
    #: When live location expires
    expires_at: Optional[datetime.datetime] = None
    #: True if live location is expired
    is_expired: Optional[bool] = None

    @classmethod
    def _from_pull(cls, data):
        return cls(
            id=data["id"],
            latitude=data["coordinate"]["latitude"] / (10 ** 8)
            if not data.get("stopReason")
            else None,
            longitude=data["coordinate"]["longitude"] / (10 ** 8)
            if not data.get("stopReason")
            else None,
            name=data.get("locationTitle"),
            expires_at=_util.millis_to_datetime(data["expirationTime"]),
            is_expired=bool(data.get("stopReason")),
        )

    @classmethod
    def _from_graphql(cls, data):
        target = data["target"]

        image = None
        media = data.get("media")
        if media and media.get("image"):
            image = Image._from_uri(media["image"])

        return cls(
            id=int(target["live_location_id"]),
            latitude=target["coordinate"]["latitude"]
            if target.get("coordinate")
            else None,
            longitude=target["coordinate"]["longitude"]
            if target.get("coordinate")
            else None,
            image=image,
            url=data.get("url"),
            name=data["title_with_entities"]["text"],
            expires_at=_util.seconds_to_datetime(target.get("expiration_time")),
            is_expired=target.get("is_expired"),
        )



import attr
from .._common import kw_only
from .. import _exception, _util, _threads

from typing import Any

#: Default attrs settings for events
attrs_event = attr.s(slots=True, kw_only=kw_only, auto_attribs=True)


@attrs_event
class Event:
    """Base class for all events."""

    @staticmethod
    def _get_thread(session, data):
        # TODO: Handle pages? Is it even possible?
        key = data["threadKey"]

        if "threadFbId" in key:
            return _threads.Group(session=session, id=str(key["threadFbId"]))
        elif "otherUserFbId" in key:
            return _threads.User(session=session, id=str(key["otherUserFbId"]))
        raise _exception.ParseError("Could not find thread data", data=data)


@attrs_event
class UnknownEvent(Event):
    """Represent an unknown event."""

    #: Some data describing the unknown event's origin
    source: str
    #: The unknown data. This cannot be relied on, it's only for debugging purposes.
    data: Any

    @classmethod
    def _parse(cls, session, data):
        raise NotImplementedError


@attrs_event
class ThreadEvent(Event):
    """Represent an event that was done by a user/page in a thread."""

    #: The person who did the action
    author: "_threads.User"  # Or Union[User, Page]?
    #: Thread that the action was done in
    thread: "_threads.ThreadABC"

    @classmethod
    def _parse_metadata(cls, session, data):
        metadata = data["messageMetadata"]
        author = _threads.User(session=session, id=metadata["actorFbId"])
        thread = cls._get_thread(session, metadata)
        at = _util.millis_to_datetime(int(metadata["timestamp"]))
        return author, thread, at

    @classmethod
    def _parse_fetch(cls, session, data):
        author = _threads.User(session=session, id=data["message_sender"]["id"])
        at = _util.millis_to_datetime(int(data["timestamp_precise"]))
        return author, at



import datetime
import json
import time
import random
import urllib.parse

from ._common import log
from . import _exception

from typing import Iterable, Optional, Any, Mapping, Sequence


def int_or_none(inp: Any) -> Optional[int]:
    try:
        return int(inp)
    except Exception:
        return None


def get_limits(limit: Optional[int], max_limit: int) -> Iterable[int]:
    """Helper that generates limits based on a max limit."""
    if limit is None:
        # Generate infinite items
        while True:
            yield max_limit

    if limit < 0:
        raise ValueError("Limit cannot be negative")

    # Generate n items
    yield from [max_limit] * (limit // max_limit)

    remainder = limit % max_limit
    if remainder:
        yield remainder


def json_minimal(data: Any) -> str:
    """Get JSON data in minimal form."""
    return json.dumps(data, separators=(",", ":"))


def strip_json_cruft(text: str) -> str:
    """Removes `for(;;);` (and other cruft) that preceeds JSON responses."""
    try:
        return text[text.index("{") :]
    except ValueError as e:
        raise _exception.ParseError("No JSON object found", data=text) from e


def parse_json(text: str) -> Any:
    try:
        return json.loads(text)
    except ValueError as e:
        raise _exception.ParseError("Error while parsing JSON", data=text) from e


def generate_offline_threading_id():
    ret = datetime_to_millis(now())
    value = int(random.random() * 4294967295)
    string = ("0000000000000000000000" + format(value, "b"))[-22:]
    msgs = format(ret, "b") + string
    return str(int(msgs, 2))


def remove_version_from_module(module):
    return module.split("@", 1)[0]


def get_jsmods_require(require) -> Mapping[str, Sequence[Any]]:
    rtn = {}
    for item in require:
        if len(item) == 1:
            (module,) = item
            rtn[remove_version_from_module(module)] = []
            continue
        module, method, requirements, arguments = item
        method = "{}.{}".format(remove_version_from_module(module), method)
        rtn[method] = arguments
    return rtn


def get_jsmods_define(define) -> Mapping[str, Mapping[str, Any]]:
    rtn = {}
    for item in define:
        module, requirements, data, _ = item
        rtn[module] = data
    return rtn


def mimetype_to_key(mimetype: str) -> str:
    if not mimetype:
        return "file_id"
    if mimetype == "image/gif":
        return "gif_id"
    x = mimetype.split("/")
    if x[0] in ["video", "image", "audio"]:
        return "%s_id" % x[0]
    return "file_id"


def get_url_parameter(url: str, param: str) -> Optional[str]:
    params = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
    if not params.get(param):
        return None
    return params[param][0]


def seconds_to_datetime(timestamp_in_seconds: float) -> datetime.datetime:
    """Convert an UTC timestamp to a timezone-aware datetime object."""
    # `.utcfromtimestamp` will return a "naive" datetime object, which is why we use the
    # following:
    return datetime.datetime.fromtimestamp(
        timestamp_in_seconds, tz=datetime.timezone.utc
    )


def millis_to_datetime(timestamp_in_milliseconds: int) -> datetime.datetime:
    """Convert an UTC timestamp, in milliseconds, to a timezone-aware datetime."""
    return seconds_to_datetime(timestamp_in_milliseconds / 1000)


def datetime_to_seconds(dt: datetime.datetime) -> int:
    """Convert a datetime to an UTC timestamp.

    Naive datetime objects are presumed to represent time in the system timezone.

    The returned seconds will be rounded to the nearest whole number.
    """
    # We could've implemented some fancy "convert naive timezones to UTC" logic, but
    # it's not really worth the effort.
    return round(dt.timestamp())


def datetime_to_millis(dt: datetime.datetime) -> int:
    """Convert a datetime to an UTC timestamp, in milliseconds.

    Naive datetime objects are presumed to represent time in the system timezone.

    The returned milliseconds will be rounded to the nearest whole number.
    """
    return round(dt.timestamp() * 1000)


def seconds_to_timedelta(seconds: float) -> datetime.timedelta:
    """Convert seconds to a timedelta."""
    return datetime.timedelta(seconds=seconds)


def millis_to_timedelta(milliseconds: int) -> datetime.timedelta:
    """Convert a duration (in milliseconds) to a timedelta object."""
    return datetime.timedelta(milliseconds=milliseconds)


def timedelta_to_seconds(td: datetime.timedelta) -> int:
    """Convert a timedelta to seconds.

    The returned seconds will be rounded to the nearest whole number.
    """
    return round(td.total_seconds())


def now() -> datetime.datetime:
    """The current time.

    Similar to datetime.datetime.now(), but returns a non-naive datetime.
    """
    return datetime.datetime.now(tz=datetime.timezone.utc)



import sys
import attr
import logging

log = logging.getLogger("fbchat")

# Enable kw_only if the python version supports it
kw_only = sys.version_info[:2] > (3, 5)

#: Default attrs settings for classes
attrs_default = attr.s(frozen=True, slots=True, kw_only=kw_only, auto_attribs=True)



import attr
from . import Attachment
from .._common import attrs_default

from typing import Any, Optional


@attrs_default
class QuickReply:
    """Represents a quick reply."""

    #: Payload of the quick reply
    payload: Any = None
    #: External payload for responses
    external_payload: Any = None
    #: Additional data
    data: Any = None
    #: Whether it's a response for a quick reply
    is_response: bool = False


@attrs_default
class QuickReplyText(QuickReply):
    """Represents a text quick reply."""

    #: Title of the quick reply
    title: Optional[str] = None
    #: URL of the quick reply image
    image_url: Optional[str] = None
    #: Type of the quick reply
    _type = "text"


@attrs_default
class QuickReplyLocation(QuickReply):
    """Represents a location quick reply (Doesn't work on mobile)."""

    #: Type of the quick reply
    _type = "location"


@attrs_default
class QuickReplyPhoneNumber(QuickReply):
    """Represents a phone number quick reply (Doesn't work on mobile)."""

    #: URL of the quick reply image
    image_url: Optional[str] = None
    #: Type of the quick reply
    _type = "user_phone_number"


@attrs_default
class QuickReplyEmail(QuickReply):
    """Represents an email quick reply (Doesn't work on mobile)."""

    #: URL of the quick reply image
    image_url: Optional[str] = None
    #: Type of the quick reply
    _type = "user_email"


def graphql_to_quick_reply(q, is_response=False):
    data = dict()
    _type = q.get("content_type").lower()
    if q.get("payload"):
        data["payload"] = q["payload"]
    if q.get("data"):
        data["data"] = q["data"]
    if q.get("image_url") and _type is not QuickReplyLocation._type:
        data["image_url"] = q["image_url"]
    data["is_response"] = is_response
    if _type == QuickReplyText._type:
        if q.get("title") is not None:
            data["title"] = q["title"]
        rtn = QuickReplyText(**data)
    elif _type == QuickReplyLocation._type:
        rtn = QuickReplyLocation(**data)
    elif _type == QuickReplyPhoneNumber._type:
        rtn = QuickReplyPhoneNumber(**data)
    elif _type == QuickReplyEmail._type:
        rtn = QuickReplyEmail(**data)
    return rtn



import attr
import datetime
from ._common import attrs_event, Event, UnknownEvent, ThreadEvent
from ._client_payload import *
from ._delta_class import *
from ._delta_type import *

from .. import _exception, _threads, _models

from typing import Mapping


@attrs_event
class Typing(ThreadEvent):
    """Somebody started/stopped typing in a thread."""

    #: ``True`` if the user started typing, ``False`` if they stopped
    status: bool

    @classmethod
    def _parse_orca(cls, session, data):
        author = _threads.User(session=session, id=str(data["sender_fbid"]))
        status = data["state"] == 1
        return cls(author=author, thread=author, status=status)

    @classmethod
    def _parse_thread_typing(cls, session, data):
        author = _threads.User(session=session, id=str(data["sender_fbid"]))
        thread = _threads.Group(session=session, id=str(data["thread"]))
        status = data["state"] == 1
        return cls(author=author, thread=thread, status=status)


@attrs_event
class FriendRequest(Event):
    """Somebody sent a friend request."""

    #: The user that sent the request
    author: "_threads.User"

    @classmethod
    def _parse(cls, session, data):
        author = _threads.User(session=session, id=str(data["from"]))
        return cls(author=author)


@attrs_event
class Presence(Event):
    """The list of active statuses was updated.

    Chat online presence update.
    """

    # TODO: Document this better!

    #: User ids mapped to their active status
    statuses: Mapping[str, "_models.ActiveStatus"]
    #: ``True`` if the list is fully updated and ``False`` if it's partially updated
    full: bool

    @classmethod
    def _parse(cls, session, data):
        statuses = {
            str(d["u"]): _models.ActiveStatus._from_orca_presence(d)
            for d in data["list"]
        }
        return cls(statuses=statuses, full=data["list_type"] == "full")


@attrs_event
class Connect(Event):
    """The client was connected to Facebook.

    This is not guaranteed to be triggered the same amount of times `Disconnect`!
    """


@attrs_event
class Resync(Event):
    """The sequence ID was reset in _handle_ms."""


@attrs_event
class Disconnect(Event):
    """The client lost the connection to Facebook.

    This is not guaranteed to be triggered the same amount of times `Connect`!
    """

    #: The reason / error string for the disconnect
    reason: str


def parse_events(session, topic, data):
    # See Mqtt._configure_connect_options for information about these topics
    try:
        if topic == "/t_ms":
            # `deltas` will always be available, since we're filtering out the things
            # that don't have it earlier in the MQTT listener
            for delta in data["deltas"]:
                if delta["class"] == "ClientPayload":
                    yield from parse_client_payloads(session, delta)
                    continue
                try:
                    event = parse_delta(session, delta)
                    if event:  # Skip `None`
                        yield event
                except _exception.ParseError:
                    raise
                except Exception as e:
                    raise _exception.ParseError(
                        "Error parsing delta", data=delta
                    ) from e

        elif topic == "/thread_typing":
            yield Typing._parse_thread_typing(session, data)

        elif topic == "/orca_typing_notifications":
            yield Typing._parse_orca(session, data)

        elif topic == "/legacy_web":
            if data["type"] == "jewel_requests_add":
                yield FriendRequest._parse(session, data)
            else:
                yield UnknownEvent(source="/legacy_web", data=data)

        elif topic == "/orca_presence":
            yield Presence._parse(session, data)

        else:
            yield UnknownEvent(source=topic, data=data)
    except _exception.ParseError:
        raise
    except Exception as e:
        raise _exception.ParseError(
            "Error parsing MQTT topic {}".format(topic), data=data
        ) from e



import attr
import datetime
from ._abc import ThreadABC
from .._common import log, attrs_default
from .. import _util, _session, _models

from typing import Optional


GENDERS = {
    # For standard requests
    0: "unknown",
    1: "female_singular",
    2: "male_singular",
    3: "female_singular_guess",
    4: "male_singular_guess",
    5: "mixed",
    6: "neuter_singular",
    7: "unknown_singular",
    8: "female_plural",
    9: "male_plural",
    10: "neuter_plural",
    11: "unknown_plural",
    # For graphql requests
    "UNKNOWN": "unknown",
    "FEMALE": "female_singular",
    "MALE": "male_singular",
    # '': 'female_singular_guess',
    # '': 'male_singular_guess',
    # '': 'mixed',
    "NEUTER": "neuter_singular",
    # '': 'unknown_singular',
    # '': 'female_plural',
    # '': 'male_plural',
    # '': 'neuter_plural',
    # '': 'unknown_plural',
}


@attrs_default
class User(ThreadABC):
    """Represents a Facebook user. Implements `ThreadABC`.

    Example:
        >>> user = fbchat.User(session=session, id="1234")
    """

    #: The session to use when making requests.
    session: _session.Session
    #: The user's unique identifier.
    id: str = attr.ib(converter=str)

    def _to_send_data(self):
        return {
            "other_user_fbid": self.id,
            # The entry below is to support .wave
            "specific_to_list[0]": "fbid:{}".format(self.id),
        }

    def _copy(self) -> "User":
        return User(session=self.session, id=self.id)

    async def confirm_friend_request(self):
        """Confirm a friend request, adding the user to your friend list.

        Example:
            >>> user.confirm_friend_request()
        """
        data = {"to_friend": self.id, "action": "confirm"}
        j = await self.session._payload_post("/ajax/add_friend/action.php?dpr=1", data)

    async def remove_friend(self):
        """Remove the user from the client's friend list.

        Example:
            >>> user.remove_friend()
        """
        data = {"uid": self.id}
        j = await self.session._payload_post("/ajax/profile/removefriendconfirm.php", data)

    async def block(self):
        """Block messages from the user.

        Example:
            >>> user.block()
        """
        data = {"fbid": self.id}
        j = await self.session._payload_post("/messaging/block_messages/?dpr=1", data)

    async def unblock(self):
        """Unblock a previously blocked user.

        Example:
            >>> user.unblock()
        """
        data = {"fbid": self.id}
        j = await self.session._payload_post("/messaging/unblock_messages/?dpr=1", data)


@attrs_default
class UserData(User):
    """Represents data about a Facebook user.

    Inherits `User`, and implements `ThreadABC`.
    """

    #: The user's picture
    photo: _models.Image
    #: The name of the user
    name: str
    #: Whether the user and the client are friends
    is_friend: bool
    #: The users first name
    first_name: str
    #: The users last name
    last_name: Optional[str] = None
    #: When the thread was last active / when the last message was sent
    last_active: Optional[datetime.datetime] = None
    #: Number of messages in the thread
    message_count: Optional[int] = None
    #: Set `Plan`
    plan: Optional[_models.PlanData] = None
    #: The profile URL. ``None`` for Messenger-only users
    url: Optional[str] = None
    #: The user's gender
    gender: Optional[str] = None
    #: From 0 to 1. How close the client is to the user
    affinity: Optional[float] = None
    #: The user's nickname
    nickname: Optional[str] = None
    #: The clients nickname, as seen by the user
    own_nickname: Optional[str] = None
    #: The message color
    color: Optional[str] = None
    #: The default emoji
    emoji: Optional[str] = None

    @staticmethod
    def _get_other_user(data):
        (user,) = (
            node["messaging_actor"]
            for node in data["all_participants"]["nodes"]
            if node["messaging_actor"]["id"] == data["thread_key"]["other_user_id"]
        )
        return user

    @classmethod
    def _from_graphql(cls, session, data):
        c_info = cls._parse_customization_info(data)

        plan = None
        if data.get("event_reminders") and data["event_reminders"].get("nodes"):
            plan = _models.PlanData._from_graphql(
                session, data["event_reminders"]["nodes"][0]
            )

        return cls(
            session=session,
            id=data["id"],
            url=data["url"],
            first_name=data["first_name"],
            last_name=data.get("last_name"),
            is_friend=data["is_viewer_friend"],
            gender=GENDERS.get(data["gender"]),
            affinity=data.get("viewer_affinity"),
            nickname=c_info.get("nickname"),
            color=c_info["color"],
            emoji=c_info["emoji"],
            own_nickname=c_info.get("own_nickname"),
            photo=_models.Image._from_uri(data["profile_picture"]),
            name=data["name"],
            message_count=data.get("messages_count"),
            plan=plan,
        )

    @classmethod
    def _from_thread_fetch(cls, session, data):
        user = cls._get_other_user(data)
        if user["__typename"] != "User":
            # TODO: Add Page._from_thread_fetch, and parse it there
            log.warning("Tried to parse %s as a user.", user["__typename"])
            return None

        c_info = cls._parse_customization_info(data)

        plan = None
        if data["event_reminders"]["nodes"]:
            plan = _models.PlanData._from_graphql(
                session, data["event_reminders"]["nodes"][0]
            )

        return cls(
            session=session,
            id=user["id"],
            url=user["url"],
            name=user["name"],
            first_name=user["short_name"],
            is_friend=user["is_viewer_friend"],
            gender=GENDERS.get(user["gender"]),
            nickname=c_info.get("nickname"),
            color=c_info["color"],
            emoji=c_info["emoji"],
            own_nickname=c_info.get("own_nickname"),
            photo=_models.Image._from_uri(user["big_image_src"]),
            message_count=data["messages_count"],
            last_active=_util.millis_to_datetime(int(data["updated_time_precise"])),
            plan=plan,
        )

    @classmethod
    def _from_all_fetch(cls, session, data):
        return cls(
            session=session,
            id=data["id"],
            first_name=data["firstName"],
            url=data["uri"],
            photo=_models.Image(url=data["thumbSrc"]),
            name=data["name"],
            is_friend=data["is_friend"],
            gender=GENDERS.get(data["gender"]),
        )



import attr
import datetime
from . import Image, Attachment
from .._common import attrs_default
from .. import _util

from typing import Set, Optional


@attrs_default
class FileAttachment(Attachment):
    """Represents a file that has been sent as a Facebook attachment."""

    #: URL where you can download the file
    url: Optional[str] = None
    #: Size of the file in bytes
    size: Optional[int] = None
    #: Name of the file
    name: Optional[str] = None
    #: Whether Facebook determines that this file may be harmful
    is_malicious: Optional[bool] = None

    @classmethod
    def _from_graphql(cls, data, size=None):
        return cls(
            url=data.get("url"),
            size=size,
            name=data.get("filename"),
            is_malicious=data.get("is_malicious"),
            id=data.get("message_file_fbid"),
        )


@attrs_default
class AudioAttachment(Attachment):
    """Represents an audio file that has been sent as a Facebook attachment."""

    #: Name of the file
    filename: Optional[str] = None
    #: URL of the audio file
    url: Optional[str] = None
    #: Duration of the audio clip
    duration: Optional[datetime.timedelta] = None
    #: Audio type
    audio_type: Optional[str] = None

    @classmethod
    def _from_graphql(cls, data):
        return cls(
            filename=data.get("filename"),
            url=data.get("playable_url"),
            duration=_util.millis_to_timedelta(data.get("playable_duration_in_ms")),
            audio_type=data.get("audio_type"),
        )


@attrs_default
class ImageAttachment(Attachment):
    """Represents an image that has been sent as a Facebook attachment.

    To retrieve the full image URL, use: `Client.fetch_image_url`, and pass it the id of
    the image attachment.
    """

    #: The extension of the original image (e.g. ``png``)
    original_extension: Optional[str] = None
    #: Width of original image
    width: Optional[int] = attr.ib(default=None, converter=_util.int_or_none)
    #: Height of original image
    height: Optional[int] = attr.ib(default=None, converter=_util.int_or_none)
    #: Whether the image is animated
    is_animated: Optional[bool] = None
    #: A set, containing variously sized / various types of previews of the image
    previews: Set[Image] = attr.ib(factory=set)

    @classmethod
    def _from_graphql(cls, data):
        previews = {
            Image._from_uri_or_none(data.get("thumbnail")),
            Image._from_uri_or_none(data.get("preview") or data.get("preview_image")),
            Image._from_uri_or_none(data.get("large_preview")),
            Image._from_uri_or_none(data.get("animated_image")),
        }

        return cls(
            original_extension=data.get("original_extension")
            or (data["filename"].split("-")[0] if data.get("filename") else None),
            width=data.get("original_dimensions", {}).get("width"),
            height=data.get("original_dimensions", {}).get("height"),
            is_animated=data["__typename"] == "MessageAnimatedImage",
            previews={p for p in previews if p},
            id=data.get("legacy_attachment_id"),
        )

    @classmethod
    def _from_list(cls, data):
        previews = {
            Image._from_uri_or_none(data["image"]),
            Image._from_uri(data["image1"]),
            Image._from_uri(data["image2"]),
        }

        return cls(
            width=data["original_dimensions"].get("x"),
            height=data["original_dimensions"].get("y"),
            previews={p for p in previews if p},
            id=data["legacy_attachment_id"],
        )


@attrs_default
class VideoAttachment(Attachment):
    """Represents a video that has been sent as a Facebook attachment."""

    #: Size of the original video in bytes
    size: Optional[int] = None
    #: Width of original video
    width: Optional[int] = None
    #: Height of original video
    height: Optional[int] = None
    #: Length of video
    duration: Optional[datetime.timedelta] = None
    #: URL to very compressed preview video
    preview_url: Optional[str] = None
    #: A set, containing variously sized previews of the video
    previews: Set[Image] = attr.ib(factory=set)

    @classmethod
    def _from_graphql(cls, data, size=None):
        previews = {
            Image._from_uri_or_none(data.get("chat_image")),
            Image._from_uri_or_none(data.get("inbox_image")),
            Image._from_uri_or_none(data.get("large_image")),
        }

        return cls(
            size=size,
            width=data.get("original_dimensions", {}).get("width"),
            height=data.get("original_dimensions", {}).get("height"),
            duration=_util.millis_to_timedelta(data.get("playable_duration_in_ms")),
            preview_url=data.get("playable_url"),
            previews={p for p in previews if p},
            id=data.get("legacy_attachment_id"),
        )

    @classmethod
    def _from_subattachment(cls, data):
        media = data["media"]
        image = Image._from_uri_or_none(media.get("image"))

        return cls(
            duration=_util.millis_to_timedelta(media.get("playable_duration_in_ms")),
            preview_url=media.get("playable_url"),
            previews={image} if image else {},
            id=data["target"].get("video_id"),
        )

    @classmethod
    def _from_list(cls, data):
        previews = {
            Image._from_uri(data["image"]),
            Image._from_uri(data["image1"]),
            Image._from_uri(data["image2"]),
        }

        return cls(
            width=data["original_dimensions"].get("x"),
            height=data["original_dimensions"].get("y"),
            previews=previews,
            id=data["legacy_attachment_id"],
        )


def graphql_to_attachment(data, size=None):
    _type = data["__typename"]
    if _type in ["MessageImage", "MessageAnimatedImage"]:
        return ImageAttachment._from_graphql(data)
    elif _type == "MessageVideo":
        return VideoAttachment._from_graphql(data, size=size)
    elif _type == "MessageAudio":
        return AudioAttachment._from_graphql(data)
    elif _type == "MessageFile":
        return FileAttachment._from_graphql(data, size=size)

    return Attachment(id=data.get("legacy_attachment_id"))


def graphql_to_subattachment(data):
    target = data.get("target")
    type_ = target.get("__typename") if target else None

    if type_ == "Video":
        return VideoAttachment._from_subattachment(data)

    return None
